{
=============================================================================
| THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  |
| AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    |
| IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   |
| ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR  |
| ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL       |
| DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR   |
| SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER   |
| CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT           |
| LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY    |
| OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  |
| DAMAGE.                                                                      |
=============================================================================
   Author:  D.J.C. Blijker.
   Date:    23-4-2003
   version: 1.0
   e-mail:  c.blijker@wxs.nl
=============================================================================
  This unit will work in Delphi as well as in Kylix!!!
=============================================================================

 Main target of this unit is to supply applications with an easy-to-use and
 consistent means of finding the exact network-address, port and protocol
 of (self implemented) network services.

 This protocol works in tight cooperation with the ServiceBroker (a windows
 example ServiceBroker application is enclosed.) and the THxServiceSeeker
 class.

 Usage:

 Prerequisites:
 -Make sure a ServiceBroker is running somewhere on your local network.
 -Make sure the 'service' you want to make available is searched by an instance
  of a THxServiceSeeker-class

 -Instantiate a ServiceProvider object.
 -Supply the netmask (if it differs from 255.255.255.0)
 -Call ProvideService with the appropriate parameters.
 -You can, if necessary, revoke a service by calling RevokeService()
  Mind you this is automatically called when the class is destroyed!
==============================================================================
History:

  version: 1.0  Initial release.
==============================================================================
}
unit ServiceProviderLayer;

interface

uses
{$IFDEF LINUX}
  QExtCtrls,
{$ELSE}
  extctrls,
{$ENDIF}
  SysUtils, HxTcpIpcUdp, HxDataExchange, ServiceBroker_Shared;

Type
  TOnServiceAck = procedure( sServiceName: string;
                             eProtocol: THxProtocol )of object;
  TOnNoServiceBroker = procedure of object;

  TOnServiceBrokerFound = procedure( sAddress, sPort: string )of object;

  THxServiceProvider = class
  private
    bServiceBrokerFound: boolean;
    bMessageGiven: boolean;
    tmr: TTimer;
    udp: THxUdp;
    de: THxDataExchange;
    lstServices: THxServiceLongList;
    sNetmask: string;
    FOnServiceAck: TOnServiceAck;
    FOnNoServiceBroker: TOnNoServiceBroker;
    FOnServiceBrokerFound: TOnServiceBrokerFound;
    procedure SetNetmask( sNM: string );
    procedure HandleTimer( Sender: TObject );
    procedure SeekService( sNameOfService: string; eProtocol: THxProtocol );
    procedure HandleDataAvailable( ptr: pchar; iSize: integer;
                                   sRemoteIP, sPort: string );
    procedure SendServiceInfo( pServInfo: PHxServiceLong; iMessNr: integer );
    procedure HandleServiceAck(ptr: pchar; iSize: integer);
    procedure HandleServiceAns(ptr: pchar; iSize: integer);
    function  GetServiceInfo( sNameOfService,
                              sAddressOfService,
                              sPort: string;
                              eProtocol: THxProtocol ): PHxServiceLong;
    procedure RevokeAllServices;
  public
    constructor Create;
    destructor Destroy; override;
    procedure ProvideService( sServName, sServPort: string;
                              eProtocol: THxProtocol );
    procedure RevokeService( sServName, sServPort: string;
                             eProtocol: THxProtocol );
    property NetMask: string read sNetMask write SetNetmask;
    property OnServiceAck: TOnServiceAck read FOnServiceAck
             write FOnServiceAck;
    property OnNoServiceBroker: TOnNoServiceBroker read FOnNoServiceBroker
             write FOnNoServiceBroker;
    property OnServiceBrokerFound: TOnServiceBrokerFound
             read FOnServiceBrokerFound write FOnServiceBrokerFound;
  end;

implementation

{ THxServiceProvider }

constructor THxServiceProvider.Create;
begin
  bServiceBrokerFound := False;
  bMessageGiven := False;

// For this service provider to work we need a valid netmask.!!
  NetMask := '255.255.255.0'; // Best guess..

  de := THxDataExchange.Create;

  tmr := TTimer.Create( nil );
  tmr.Interval := 10000;      // Refresh list-of-services at the serviceBroker
  tmr.OnTimer := HandleTimer; // every 10 seconds..
  tmr.Enabled := False;

  udp := THxUdp.Create;
  udp.ListeningPort := ANY_FREE_PORT;
  udp.DestinationAddress := GetBroadcastAddress( Netmask );
  udp.DestinationPort := SERVICEBROKER_PORT;
  udp.OnDataAvailable := HandleDataAvailable;

  lstServices := THxServiceLongList.Create;
end;

destructor THxServiceProvider.Destroy;
begin
  RevokeAllServices;

  udp.Free;
  tmr.Free;
  lstServices.Free;

  inherited;
end;

procedure THxServiceProvider.HandleTimer( Sender: TObject );
begin
  if not( bServiceBrokerFound )then
  begin
    bMessageGiven := False;
    if Assigned( FOnNoServiceBroker )then
      FOnNoServiceBroker;
  end;

  bServiceBrokerFound := False;
  SeekService( SERVICEBROKER_SERVICE, pcUdp );
end;

procedure THxServiceProvider.RevokeAllServices;
var
  i: integer;
begin
  i := lstServices.Count -1;
  while i >= 0 do
  begin
    SendServiceInfo( lstServices[i], REVOKE_SERVICE );
    lstServices.Delete( i );
    Dec( i );
  end;
end;

procedure THxServiceProvider.SeekService( sNameOfService: string;
                                          eProtocol: THxProtocol );
var
  ptr: pchar;
  iSize: integer;
begin
  de.StartCreatingMessage;
  de.AddString( SAP_HEADER );
  de.AddInteger( SERVICE_REQ );
  de.AddString( sNameOfService );
  de.AddInteger( integer( eProtocol ));
  de.GetCreatedMessage( ptr, iSize );

  udp.DestinationAddress := GetBroadcastAddress( sNetMask );
  udp.DestinationPort := SERVICEBROKER_PORT;
  udp.SendData( ptr, iSize );
end;

procedure THxServiceProvider.SetNetMask( sNM: string );
var
  s: string;
begin
  s := ValidateIpAddress( sNM );
  if s <> '' then
    sNetMask := s
  else
    raise Exception.Create( sNM+' is not a valid netmask!');
end;

procedure THxServiceProvider.HandleDataAvailable( ptr: pchar; iSize: integer;
                                                sRemoteIP, sPort: string );
var
  sSAPHeader: string;
  iMessNr: integer;
begin
  try
    ptr := de.ReadString( ptr, sSAPHeader, iSize );
    if sSAPHeader <> SAP_HEADER then
      exit; // ignore 'strange' messages..
    ptr := de.ReadInteger( ptr, iMessNr, iSize );
    case iMessNr of
      INVALID_NUMBER: exit; // ignore invalid messages..
      ACK_SERVICE_ADDRESS: HandleServiceAck( ptr, iSize );
      SERVICE_ANS: HandleServiceAns( ptr, iSize );
    else
      exit; // ignore not supported messages..
    end;
  except
    // ignore also..
  end;
end;

procedure THxServiceProvider.HandleServiceAck( ptr: pchar; iSize: integer );
var
  sServiceName, sServicePort, sServiceAddress: string;
  iProtocol, i: integer;
  eProtocol: THxProtocol;
begin
  ptr := de.ReadString( ptr, sServiceName, iSize );
  if iSize = 0 then
    exit;

  ptr := de.ReadString( ptr, sServiceAddress, iSize );
  if iSize = 0 then
    exit;

  ptr := de.ReadString( ptr, sServicePort, iSize );
  if iSize = 0 then
    exit;

  de.ReadInteger( ptr, iProtocol, iSize );
  if( iSize <> 0 )or   // Should be ready now..
  ( iProtocol < 0 )or
  ( iProtocol > integer( High( THxProtocol )) )then
    exit;

  eProtocol := THxProtocol( iProtocol );

  for i := 0 to lstServices.Count-1 do
  begin
    if( lstServices[i]^.sServiceName = sServiceName )and
    ( lstServices[i]^.sPort = sServicePort )and
    ( lstServices[i]^.eProtocolType = eProtocol )and
    ( lstServices[i]^.sAddress = sServiceAddress )and
    ( not( lstServices[i]^.bServAck ))then
    begin
      lstServices[i]^.bServAck := True;

      if Assigned( FOnServiceAck )then
        FOnServiceAck( sServiceName, eProtocol );

      break;
    end;
  end;
end;

procedure THxServiceProvider.HandleServiceAns( ptr: pchar; iSize: integer );
var
  sServiceName, sServicePort, sServiceAddress: string;
  iProtocol, i: integer;
  eProtocol: THxProtocol;
begin
  ptr := de.ReadString( ptr, sServiceName, iSize );
  if iSize = 0 then
    exit;

  ptr := de.ReadString( ptr, sServiceAddress, iSize );
  if iSize = 0 then
    exit;

  ptr := de.ReadString( ptr, sServicePort, iSize );
  if iSize = 0 then
    exit;

  de.ReadInteger( ptr, iProtocol, iSize );
  if( iSize <> 0 )or   // Should be ready now..
  ( iProtocol < 0 )or
  ( iProtocol > integer( High( THxProtocol )) )then
    exit;

  bServiceBrokerFound := True;

  eProtocol := THxProtocol( iProtocol );
  if eProtocol = pcUdp then
    for i := 0 to lstServices.Count-1 do
      SendServiceInfo( lstServices[i], NEW_SERVICE_ADDRESS );

  if( not( bMessageGiven ))and( Assigned( FOnServiceBrokerFound ))then
  begin
    FOnServiceBrokerFound( udp.LastSendersAddress, udp.LastSendersPort );
    bMessageGiven := True;
  end;
end;

procedure THxServiceProvider.SendServiceInfo( pServInfo: PHxServiceLong;
                                            iMessNr: integer );
var
  ptr: pchar;
  iSize: integer;
begin
  if not( bServiceBrokerFound )then
    exit;

  de.StartCreatingMessage;
  de.AddString( SAP_HEADER );
  de.AddInteger( iMessNr );
  de.AddString( pServInfo^.sServiceName );
  de.AddString( pServInfo^.sAddress );
  de.AddString( pServInfo^.sPort );
  de.AddInteger( integer( pServInfo^.eProtocolType ));
  de.GetCreatedMessage( ptr, iSize );

  udp.Reply( ptr, iSize );
end;


procedure THxServiceProvider.ProvideService( sServName, sServPort: string;
                                           eProtocol: THxProtocol );
var
  pServ: PHxServiceLong;
begin
  New( pServ );
  lstServices.Add( pServ );
  pServ^.sServiceName := sServName;
  pServ^.sAddress := GetLocalAddress;
  pServ^.sPort := sServPort;
  pServ^.eProtocolType := eProtocol;
  pServ^.bServAck := False;

  if bServiceBrokerFound then
    SendServiceInfo( pServ, NEW_SERVICE_ADDRESS )
  else
    SeekService( SERVICEBROKER_SERVICE, pcUdp );

  tmr.Enabled := True;
end;

procedure THxServiceProvider.RevokeService( sServName, sServPort: string;
                                          eProtocol: THxProtocol );
var
  sServAddr: string;
  pServ: PHxServiceLong;
begin
  sServAddr := GetLocalAddress;
  pServ := GetServiceInfo( sServName, sServAddr, sServPort, eProtocol );
  if pServ = nil then
    exit;

  SendServiceInfo( pServ, REVOKE_SERVICE );

  lstServices.Delete( lstServices.IndexOf( pServ ));

  tmr.Enabled := lstServices.Count <> 0;
end;

function THxServiceProvider.GetServiceInfo( sNameOfService,
                                      sAddressOfService,
                                      sPort: string;
                                      eProtocol: THxProtocol ): PHxServiceLong;
var
  i: integer;
begin
  result := nil;
  for i := 0 to lstServices.Count -1 do
    if( lstServices[i]^.sServiceName = sNameOfService )and
    ( lstServices[i]^.sPort = sPort )and
    ( lstServices[i]^.eProtocolType = eProtocol )and
    ( lstServices[i]^.sAddress = sAddressOfService )then
    begin
      result := lstServices[i];
      break;
    end;
end;

end.
